#include "port_audio_device.h"

#include "audio/mixer/mixer.h"
#include "audio/mixer/mixer_track.h"
#include "audio/plugin_host/midi_sequencer.h"
#include "audio/plugin_host/plugin_host_manager.h"
#include "thread_message/thread_message_hubs.h"

int port_audio_callback(const void* Input, void* Output, unsigned long FrameCount, const PaStreamCallbackTimeInfo* TimeInfo, PaStreamCallbackFlags StatusFlags, void* UserData)
{
    auto* API = static_cast<port_audio_device*>(UserData);
    return API->stream_callback((float**)Input, static_cast<float**>(Output), FrameCount, TimeInfo, StatusFlags);
}

port_audio_device::~port_audio_device() {
    port_audio_device::destroy();
}

bool port_audio_device::init() {
    Pa_Initialize();
    return true;
}

void port_audio_device::destroy() {
    stop();
    Pa_Terminate();
}

void port_audio_device::stop() {
    Pa_StopStream(stream_);
    Pa_CloseStream(stream_);
    stream_ = nullptr;
    stop_render_thread();
    spdlog::info("port_audio stopped");
}

void port_audio_device::start() {
    try {
        open_stream(-1, Pa_GetDefaultOutputDevice(), sample_rate(), buffer_size());
    } catch (const std::exception& e) {
        spdlog::error("Init failed: {}", e.what());
    }
}

void port_audio_device::open_stream(PaDeviceIndex input_device, PaDeviceIndex output_device, double in_sample_rate, uint32_t in_buffer_size) {
    const PaStreamParameters* input_params = nullptr;
    if (input_device >= 0) {
        input_params_.device = input_device;
        input_params_.channelCount = 2;
        input_params_.sampleFormat = paFloat32 | paNonInterleaved;
        input_params_.suggestedLatency = Pa_GetDeviceInfo(input_device)->defaultLowInputLatency;
        input_params_.hostApiSpecificStreamInfo = nullptr;
        input_params = &input_params_;
    }

    output_params_.device = output_device;
    output_params_.channelCount = 2;
    output_params_.sampleFormat = paFloat32 | paNonInterleaved;
    output_params_.suggestedLatency = Pa_GetDeviceInfo(output_device)->defaultLowOutputLatency;
    output_params_.hostApiSpecificStreamInfo = nullptr;
    PaError err = Pa_OpenStream(&stream_, input_params, &output_params_, in_sample_rate, in_buffer_size, paClipOff, port_audio_callback, this);
    if (err != paNoError) {
        throw std::runtime_error(Pa_GetErrorText(err));
    }
    err = Pa_StartStream(stream_);
    if (err != paNoError) {
        throw std::runtime_error(Pa_GetErrorText(err));
    }
    start_render_thread();
}

int port_audio_device::stream_callback(float** input, float** output, unsigned long frame_count,
    const PaStreamCallbackTimeInfo* time_info, PaStreamCallbackFlags status_flags) {
    if (render_buffer_.Num() < frame_count * output_params_.channelCount)
        return paContinue;
    for (int i = 0; i < output_params_.channelCount; ++i) {
        render_buffer_.Pop(output[i], frame_count);
    }
    return paContinue;
}

bool port_audio_device::on_set_buffer_size(uint32_t in_buffer_size) {
    stop();
    render_buffer_size_ = in_buffer_size * 4;
    try {
        open_stream(input_params_.device, output_params_.device, sample_rate(), in_buffer_size);
    } catch (const std::exception& e) {
        spdlog::error("Set buffer size failed: {}", e.what());
        return false;
    }
    return true;
}

bool port_audio_device::on_set_sample_rate(double in_sample_rate) {
    stop();
    try {
        open_stream(input_params_.device, output_params_.device, in_sample_rate, buffer_size());
    } catch (const std::exception& e) {
        spdlog::error("Set sample rate failed: {}", e.what());
        return false;
    }
    return true;
}

void port_audio_device::start_render_thread() {
    render_thread_ = std::thread(&port_audio_device::render_thread, this);
    render_thread_.detach();
}

void port_audio_device::stop_render_thread() {
    if (render_thread_.joinable())
        render_thread_.join();
}

void port_audio_device::render_thread() {
    dummy_track* master = g_mixer.get_master();
    spdlog::info("port_audio render thread started");
    render_buffer_.SetCapacity(buffer_size() * 4);
    while (stream_) {
        const uint32_t frames = buffer_size();
        const double rate = sample_rate();

        const uint32_t milliseconds = 1.f / (rate / (frames / 4)) * 1000;

        g_audio_thread_hub.process_messages();
        if (render_buffer_.Num() >= render_buffer_size_) {
            std::this_thread::sleep_for(std::chrono::milliseconds(milliseconds));
            continue;
        }
        g_midi_sequencer.process(rate, frames);
        g_mixer.reset();
        g_plugin_host_manager.process(frames);
        g_mixer.process(frames);

        const auto& master_headers = master->buffer.get_headers_vector();
        for (const auto element: master_headers) {
            render_buffer_.Push(element, frames);
        }
    }
    spdlog::info("port_audio render thread stopped");
}
